#!/usr/bin/python3
# Copyright eBrevia.com 2019
# pylint: disable=missing-docstring, invalid-name
# A trivial client for doc_server
# Reads all of stdin, sends it to doc_server,
# splits, sends result to stdout
# This is equivalent to doc_split but with less overhead

import socket
import sys

from compound_split import doc_config

if len(sys.argv) > 2:
    host = sys.argv[2]
else:
    host = doc_config.DEFAULT_HOST
if len(sys.argv) > 1:
    port = int(sys.argv[1])
else:
    port = doc_config.DEFAULT_PORT

with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as client:
    client.connect((host, port))
    print("connected to server", file=sys.stderr)
    input_str = sys.stdin.read()
    print("input read", file=sys.stderr)
    input_bytes = input_str.encode()
    client.sendall(input_bytes)
    print("input sent", file=sys.stderr)
    client.shutdown(socket.SHUT_WR)
    print("shut down write side", file=sys.stderr)

    output_bytes = bytearray(b'')
    blockno = 0
    while True:
        data = client.recv(2048)
        if not data:
            break
        # print("read block %d" % (blockno, file=sys.stderr))
        blockno += 1
        output_bytes += data
    print("finished reading", file=sys.stderr)
    output_str = output_bytes.decode()
    sys.stdout.write(output_str)



#!/usr/bin/python3
"""
Estimate compound word boundaries based ngram probabilities of n characters at word endings
Take as input a file with one word (a noun) per line
Use different output files for different languages (on the command line)
Specify language on the command line (de, nl, etc.)
"""

__author__ = 'don.tuggener@gmail.com'

import re
import sys
from collections import defaultdict

if len(sys.argv) != 3:
    print("usage: char_split_train.py <wordlist> <lang>", file=sys.stderr)
    sys.exit(1)

# Dicts for counting the ngrams
end_ngrams = defaultdict(int)
start_ngrams = defaultdict(int)
in_ngrams = defaultdict(int)
all_ngrams = defaultdict(int)

# Counters and limits
c = 0  # Line counter
max_words = 10000000  # Words to consider
max_len = 20  # Maximum ngram length

# Gather counts
print('Words analyzed of max.', str(max_words))
for line in open(sys.argv[1], 'r'):

    line = line.strip().lower()
    if '-' in line:
        line = re.sub('.*-', '', line)  # Hyphen: take part following last hyphen      
    line_middle = line[1:-1]

    for n in range(3, max_len + 1):  # "Overcount" long words
        # for n in range(3, len(line)+1):   # Lower performance

        if n <= max_len:
            ngram = line[:n]  # start_grams: max_len 3-5
            start_ngrams[ngram] += 1
            all_ngrams[ngram] += 1

            ngram = line[-n:]  # end_grams: max_len 3-5
            end_ngrams[ngram] += 1
            all_ngrams[ngram] += 1

        for m in range(len(line_middle) - n + 1):  # in_grams: max_len 3-5
            ngram = line_middle[m:m + n]
            if not ngram == '':
                in_ngrams[ngram] += 1
                all_ngrams[ngram] += 1

    if c % 10000 == 0:
        sys.stderr.write('\r' + str(c))
        sys.stderr.flush()
    c += 1
    if c == max_words:
        break
sys.stderr.write('\n')

print('Calculating ngrams probabilities')
start_ngrams = {k: v / all_ngrams[k] for k, v in start_ngrams.items() if v > 1}
end_ngrams = {k: v / all_ngrams[k] for k, v in end_ngrams.items() if v > 1}
in_ngrams = {k: v / all_ngrams[k] for k, v in in_ngrams.items() if v > 1}

# Write dicts to python file
with open(sys.argv[2] + '_ngram_probs.py', 'w') as f:
    f.write('prefix=')
    f.write(str(dict(start_ngrams)))
    f.write('\n')
    f.write('infix=')
    f.write(str(dict(in_ngrams)))
    f.write('\n')
    f.write('suffix=')
    f.write(str(dict(end_ngrams)))
    f.write('\n')



#!/usr/bin/python3
# Copyright eBrevia.com 2019
"""Document-level splitting module."""

import logging
import re
import sys
import pkgutil

from compound_split import doc_config
from compound_split import char_split

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

DE_LETTER = ('ABCDEFGHIJKLMNOPQRSTUVWXYZ' +
             'abcdefghijklmnopqrstuvwxyz' +
             'ÄÖÜäöüß')
DE_WORD_PAT = re.compile(r'\b[%s]+\b' % (DE_LETTER))
GLUES = ['ens', 'en', 'er', 'es', 'e', 'n', 's']
MIDDLE_DOT = '\N{MIDDLE DOT}'

# Caches
KNOWN_WORDS = set()
ALREADY_SPLIT_WORDS = {}


def find_root(word):
    """Looks in KNOWN_WORDS; if that fails, peels off German compound glue.
       If nothing works, return None.
    """
    if word in KNOWN_WORDS:
        return word
    # might not actually be a noun
    not_noun = word.lower()
    if not_noun in KNOWN_WORDS:
        return not_noun
    # do glue processing
    for glue in GLUES:
        if word.endswith(glue):
            root_length = len(word) - len(glue)
            root = word[:root_length]
            if root in KNOWN_WORDS:
                return word
    return None


def get_best_split(word):
    """First look up the word, and if found, return it.
       Then get the best available split of a word into 2 words.
       If score of best is less than 0, no split is available.
       If no split is available, the two words are the same; return one.
    """
    root = find_root(word)
    if root:
        return [root]
    candidate = char_split.split_compound(word)[0]
    if candidate[0] <= 0:
        return [word]
    if candidate[1] == candidate[2]:
        return [candidate[1]]
    return [candidate[1], candidate[2]]


def maximal_split(word, de_dict=doc_config.DEFAULT_DICT):
    """Recursively split a single word into a list of words.
       Only the first result is split further, as CharSplit divides
       compounds into non-final and final parts.
       Try to avoid splitting words other than nouns,
       as false positives are too likely.
    """
    # This is an entry point, so load the dictionary just in case.
    load_known_words(de_dict)
    # Do not split a non-noun
    if not word[0].isupper():
        return [word]
    word_list = get_best_split(word)
    # Binary splitter was unable to split
    if len(word_list) == 1:
        return word_list
    # If split product is too short, ignore the split
    if len(word_list[0]) < 4 or len(word_list[1]) < 4:
        return [word]
    # Recursively split the non-head and prepend it to the head
    return maximal_split(word_list[0]) + [word_list[1]]


def load_known_words(de_dict=doc_config.DEFAULT_DICT, de_misc=True):
    """Load the dictionary into KNOWN_WORDS."""
    if KNOWN_WORDS:
        return  # already loaded
    if de_dict is None:
        de_dict = doc_config.DEFAULT_DICT
    load_dict(de_dict)
    if de_misc:
        load_dict(doc_config.DE_MISC_DICT)
    # print('%d known words loaded\n' % (len(KNOWN_WORDS)), file=sys.stderr)
    logger.info('%d known words loaded', len(KNOWN_WORDS))


def load_dict(de_dict):
    data = pkgutil.get_data(__name__, de_dict).decode('utf-8')
    for word in data:
        if not (word == '' or word.startswith('#')):
            KNOWN_WORDS.add(word.strip())


def maximal_split_str(word, de_dict=None):
    """Maximally split a word and return it with middle dots."""
    # This is an entry point, so load the dictionary just in case.
    load_known_words(de_dict)
    # if memoized, don't split
    try:
        return ALREADY_SPLIT_WORDS[word]
    except KeyError:
        pass
    upper_case = word[0].isupper()
    result_list = [w.lower() for w in maximal_split(word)]
    result_str = MIDDLE_DOT.join(result_list)
    if upper_case:
        result0 = result_str[0].upper()
    else:
        result0 = result_str[0]
    result = result0 + result_str[1:]
    ALREADY_SPLIT_WORDS[word] = result
    return result


def add_additional_known_words(additional_known_words):
    """Add additional known words added by the user"""
    if additional_known_words is not None:
        if not isinstance(additional_known_words, list):
            raise ValueError(
                f"additional_knwon_words should be of type list and not of type {type(additional_known_words)}")
        for word in additional_known_words:
            if not word == '':
                KNOWN_WORDS.add(word)


def doc_split(doc, de_dict=None, result_map=None, additional_known_words=None):
    """Split a whole document (a string) using the specified dictionary.
       Return the whole document with splitting dots.
    """
    if doc == '':
        return ''
    # This is an entry point, so load the dictionary just in case.
    load_known_words(de_dict)
    add_additional_known_words(additional_known_words)
    result = []
    if result_map is not None:
        result_map.clear()
    windexes = [(mobj.start(), mobj.end()) for mobj in DE_WORD_PAT.finditer(doc)]
    # Non-word before the first word (OK if empty)
    result.append(doc[:windexes[0][0]])
    # pylint: disable=consider-using-enumerate
    for i in range(0, len(windexes)):
        # Add a split word and the following non-word (OK if empty)
        (start, end) = windexes[i]
        unsplit_word = doc[start:end]
        split_word = maximal_split_str(unsplit_word)
        result.append(split_word)
        if result_map is not None and MIDDLE_DOT in split_word:
            result_map[unsplit_word] = split_word
        if i == len(windexes) - 1:
            next_start = len(doc)
        else:
            next_start = windexes[i + 1][0]
        result.append(doc[end:next_start])
    return ''.join(result)  # TODO also return all splitted words for easy checking


def main():
    """Read whole document from stdin, output in maximally split format.
       Usage: ./doc_split.py dict
    """
    if len(sys.argv) > 1:
        de_dict = sys.argv[1]
    else:
        de_dict = doc_config.DEFAULT_DICT
    input_str = sys.stdin.read()
    output_str = doc_split(input_str, de_dict)
    sys.stdout.write(output_str)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import socket
import subprocess
import time
import unittest

from compound_split.doc_split import *
from compound_split.char_split import *

TEST_WORD = 'Autobahnraststätte'
RESULT_WORD1 = 'Autobahn'
RESULT_WORD2 = 'Raststätte'

TEST_SENTENCE = \
    """Die Technik setzt sich aus dem europaweiten Mobilfunk·standard Gsm,
       der in Deutschland über das D1-Netz angeboten wird,
       und dem weltweit verfügbaren System
       von Navigationssatelliten (Gps) zusammen.
    """

RESULT_SENTENCE = \
    """Die Technik setzt sich aus dem europaweiten Mobil·funk·standard Gsm,
       der in Deutschland über das D1-Netz angeboten wird,
       und dem weltweit verfügbaren System
       von Navigations·satelliten (Gps) zusammen.
    """

RESULT_DICTIONARY = \
    """Mobilfunk\tMobil·funk
    Navigationssatelliten\tNavigations·satelliten
    """

PPORT = 30302  # Don't use production port
DPORT = 30303  # Don't use production port


class TestDeDecompound(unittest.TestCase):
    """Test German decompounder.
       WARNING: If these tests fail, that does NOT necessarily mean
       that the algorithm is broken.  It may in fact have been improved,
       in which case RESULT_SENTENCE should be changed to the new result.
    """

    def test_char_split(self):
        result = split_compound(TEST_WORD)[0]
        self.assertEqual(result[1], RESULT_WORD1)
        self.assertEqual(result[2], RESULT_WORD2)

    def test_maximal_split(self):
        self.assertEqual(maximal_split('Mobilfunkstandard'),
                         ['Mobil', 'Funk', 'Standard'])
        self.assertEqual(maximal_split('europaweiten'),
                         ['europaweiten'])

    def test_maximal_split_str(self):
        self.assertEqual(maximal_split_str('Mobilfunkstandard'),
                         'Mobil·funk·standard')
        self.assertEqual(maximal_split_str('europaweiten'),
                         'europaweiten')
        self.assertEqual(maximal_split_str('unbestimmte'),
                         'unbestimmte')

    def test_doc_split(self):
        self.assertEqual(doc_split(TEST_SENTENCE), RESULT_SENTENCE)

    def test_doc_server_plaintext(self):
        # Start server
        pid = subprocess.Popen([sys.executable,  # this Python
                                '-m',
                                'doc_server',
                                '-p',
                                str(PPORT)])
        time.sleep(5)
        # Modified version of doc_client
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as client:
            client.connect(('localhost', PPORT))
            # print("connected to server", file=sys.stderr)
            input_bytes = TEST_SENTENCE.encode()
            client.sendall(input_bytes)
            # print("input sent", file=sys.stderr)
            client.shutdown(socket.SHUT_WR)
            # print("shut down write side", file=sys.stderr)
            data = client.recv(2048)  # one block is enough
            output_str = data.decode()
            # print("finished reading", file=sys.stderr)

            # Compare
            self.assertEqual(output_str, RESULT_SENTENCE)

            # Kill server
            pid.kill()

    def test_doc_server_dict(self):
        # Start server
        pid = subprocess.Popen([sys.executable,  # this Python
                                '-m',
                                'doc_server',
                                '-d',
                                str(DPORT)])
        time.sleep(5)
        # Modified version of doc_client
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as client:
            client.connect(('localhost', DPORT))
            # print("connected to server", file=sys.stderr)
            input_bytes = TEST_SENTENCE.encode()
            client.sendall(input_bytes)
            # print("input sent", file=sys.stderr)
            client.shutdown(socket.SHUT_WR)
            # print("shut down write side", file=sys.stderr)
            data = client.recv(2048)  # one block is enough
            output_str = data.decode()
            # print("finished reading", file=sys.stderr)

            # Compare
            self.assertEqual(output_str, RESULT_DICTIONARY)

            # Kill server
            pid.kill()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3
# This file is dedicated to the public domain
# Config file shared between doc_client and doc_server

# pylint: disable=unused-variable, missing-docstring
# Default server host and port and dictionary
DEFAULT_HOST = "localhost"
DEFAULT_PORT = 60604
DEFAULT_DICT = 'dicts/de-mixed.dic'
DE_MISC_DICT = 'dicts/de-misc.dic'



#!/usr/bin/python3
# Copyright eBrevia.com 2019
"""Document splitting server."""

import logging
import socket
import sys
import resource

from compound_split import doc_config
from compound_split import doc_split

result_map = {}
dict_mode = False

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
# pylint: disable=invalid-name
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def tabular(tab_map):
    """Convert Python dict into sorted output lines "unsplit\tsplit\n"."""
    result = []
    for orig in sorted(tab_map.keys()):
        split = tab_map[orig]
        result.append(orig + '\t' + split)
    return '\n'.join(result) + '\n'


def run(client_socket, client_address, dict):
    """Reads, splits, writes."""
    input_bytes = bytearray(b'')
    blockno = 0
    while True:
        block = client_socket.recv(2048)
        # logger.info("Read block %d", blockno)
        blockno += 1
        if not block:
            break
        input_bytes += block
    # logger.info("Read %d bytes", len(input_bytes))
    input_str = input_bytes.decode()
    result_map.clear()
    output_str = doc_split.doc_split(input_str, dict, result_map)
    if dict_mode:
        output_str = tabular(result_map)
    output_bytes = output_str.encode()
    # logger.info("Split the text")
    client_socket.sendall(output_bytes)
    # logger.info("Written %d bytes", len(output_bytes))
    client_socket.shutdown(socket.SHUT_WR)
    # logger.info("Client at %s disconnecting", client_address)


def main():
    """Main server program.
       Reads the whole of standard input, splits it all,
       sends the result to standard output.
       Usage: doc_server [-p][-d] port dict
    """
    if len(sys.argv) > 3:
        dict = sys.argv[3]
    else:
        dict = doc_config.DEFAULT_DICT

    if len(sys.argv) > 2:
        port = int(sys.argv[2])
    else:
        port = doc_config.DEFAULT_PORT

    global dict_mode
    if len(sys.argv) > 1:
        dict_mode = (sys.argv[1] == '-d')

    # Load dictionaries and report memory usage while quiescent
    doc_split.load_known_words(dict)
    mem_mb = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss // 1024
    logger.info("Memory usage: %d MB", mem_mb)

    while True:
        try:
            server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server.bind(('localhost', port))
            logger.info("Server started")
            while True:
                server.listen(5)
                client, client_address = server.accept()
                run(client, client_address, dict)
        except Exception:
            # trap all errors so the server doesn't die
            logger.exception("Error in doc_server main loop:");
            server.close()


if __name__ == "__main__":
    main()



#!/usr/bin/python3
"""
Split German, Dutch, etc. compound words
"""

__author__ = 'don.tuggener@gmail.com'

import sys

from compound_split import de_ngram_probs  # trained with char_split_train.py
from compound_split import nl_ngram_probs  # trained with char_split_train.py


# import other ngram_probs files only as needed to save memory


def split_compound(word: str, lang: str = 'de'):
    """
    Return list of possible splits, best first
    :param word: Word to be split
    :return: List of all splits
    """

    if lang == 'de':
        prefix = de_ngram_probs.prefix
        infix = de_ngram_probs.infix
        suffix = de_ngram_probs.suffix
    elif lang == 'nl':
        prefix = nl_ngram_probs.prefix
        infix = nl_ngram_probs.infix
        suffix = nl_ngram_probs.suffix
    # elif other languages
    else:
        raise ValueError("No model for %s language" % (lang))

    word = word.lower()

    # If there is a hyphen in the word, return part of the word behind the last hyphen
    if '-' in word:
        hyphen_index = word.rfind('-')
        return [(1., word[:hyphen_index].title(), word[hyphen_index + 1:].title())]

    scores = []  # Score for each possible split position
    # Iterate through characters, start at forth character, go to 3rd last
    for n in range(3, len(word) - 2):

        pre_slice = cut_off_fugen_s(word[:n])

        # Start, in, and end probabilities
        pre_slice_prob = suffix.get(pre_slice, -1)
        in_slice_prob = []
        ngram = cut_off_fugen_s(word[n:])
        start_slice_prob = prefix.get(ngram, -1)

        # Extract all ngrams
        for k in range(len(word), n + 2, -1):
            # Probability of ngram in word, if high, split unlikely
            in_ngram = word[n:k]
            in_slice_prob.append(infix.get(in_ngram, 1))  # Favor ngrams not occurring within words

        in_slice_prob = min(in_slice_prob)  # Lowest, punish splitting of good ingrams
        score = start_slice_prob - in_slice_prob + pre_slice_prob
        scores.append((score, word[:n].title(), word[n:].title()))

    if not scores:
        return [(0, word.title(), word.title())]
    scores.sort(reverse=True)
    return scores


def cut_off_fugen_s(word):
    if word.endswith('ts') or word.endswith('gs') or word.endswith('ks') \
            or word.endswith('hls') or word.endswith('ns'):
        if len(word[:- 1]) > 2:
            return word[:-1]
    return word


def germanet_evaluation(print_errors: bool = False):
    """ Test on GermaNet compounds from http://www.sfs.uni-tuebingen.de/lsd/compounds.shtml """
    cases, correct = 0, 0
    for line in open('split_compounds_from_GermaNet13.0.txt', 'r').readlines()[2:]:
        cases += 1
        sys.stderr.write('\r' + str(cases))
        sys.stderr.flush()
        line = line.strip().split('\t')
        if not len(line) == 3:
            continue  # A few corrupted lines
        split_result = split_compound(line[0])
        if split_result:
            if split_result[0][2] == line[2]:
                correct += 1
            elif print_errors:
                print(line, split_result)
        if cases % 10000 == 0: print(' Accuracy (' + str(correct) + '/' + str(cases) + '): ', 100 * correct / cases)
    print(' Accuracy (' + str(correct) + '/' + str(cases) + '): ', 100 * correct / cases)


if __name__ == '__main__':
    do_eval = False
    if do_eval:
        germanet_evaluation(print_errors=False)
    for x in split_compound(sys.argv[1]):
        print('\t'.join([str(y) for y in x]))



import setuptools

setuptools.setup(
    name='compound_split',
    version='1.0.2.dev2',
    install_requires=[],
    python_requires='>=3',
    author="Don Tuggener",
    author_email="don.tuggener@gmail.com",
    maintainer="Joel Niklaus",
    maintainer_email="me@joelniklaus.ch",
    description="Splits a compound into its body and head. So far German and Dutch are supported.",
    license='GPL-3.0 License',
    long_description=open('README.md', "r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/JoelNiklaus/CompoundSplit",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ]
)
